# 发送端分块
import sys
import zlib
from io import BytesIO

from ksrpc.utils.tqdm import update_progress


async def data_sender(data, print, chunk_size=1024 * 128):
    file = sys.stderr
    print(f'加载数据: ({len(data):,} bytes) [', end='', file=file)
    data = BytesIO(data)
    chunk = data.read(chunk_size)
    i = -1
    size = 0
    while chunk:
        buf = zlib.compress(chunk)
        size += len(buf)
        yield buf
        i += 1
        update_progress(i, print, file=file)
        chunk = data.read(chunk_size)
    # 会快速的显示到此处，但实际数据还在发送中，建议与客户端接收进度一起看
    print(f'] 压缩完成 ({size:,} bytes) | 底层异步发送中，请勿立即退出！', file=file)


async def send_in_chunks(ws, data, print, chunk_size=1024 * 128):  # 32KB
    # 64KB时，某宽发出的大数据无法zlib解压，所以改成了32KB
    # 将数据分块发送
    file = sys.stderr
    print(f'发送数据: ({len(data):,} bytes) [', end='', file=file)

    j = -1
    size = 0
    for i in range(0, len(data), chunk_size):
        chunk = data[i:i + chunk_size]
        buf = zlib.compress(chunk, 6)
        size += len(buf)
        await ws.send_bytes(buf)
        await ws.send_str("\r\n")
        j += 1
        update_progress(j, print, file=file)
    # 会快速的显示到此处，但实际数据还在发送中，建议与客户端接收进度一起看
    print(f'] 压缩完成 ({size:,} bytes) | 底层异步发送中，请勿立即退出！', file=file)

    # 发送结束标志
    await ws.send_str("EOF")



import asyncio
import sys
import time
import zlib

import aiohttp
import dill as pickle

from ksrpc.connections import BaseConnection
from ksrpc.utils.chunks import send_in_chunks
from ksrpc.utils.tqdm import update_progress, muted_print





class WebSocketConnection(BaseConnection):
    """WebSocket支持长连接
    1. 比HTTP协议开销较小，延迟较低
    2. 同一连接中，请求不能并行。如需并发要分别建立连接
    """

    def __init__(self, url, username=None, password=None):
        super().__init__(url, username, password)
        self._ws = None
        self._lock = asyncio.Lock()
        self._timeout = aiohttp.ClientTimeout(total=60)
        self._session = aiohttp.ClientSession(auth=self._auth, timeout=self._timeout)

    async def __aenter__(self):
        """异步async with"""
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        """异步async with"""
        async with self._lock:
            if self._ws:
                await self._ws.__aexit__(exc_type, exc_value, traceback)
            if self._session:
                await self._session.__aexit__(exc_type, exc_value, traceback)

    async def connect(self):
        async with self._lock:
            if self._ws is not None:
                return
            self._ws = await self._session.ws_connect(
                self._url.format(time=time.time()),
                # proxy="http://192.168.31.33:9000",
            ).__aenter__()

    async def reset(self):
        async with self._lock:
            if self._ws is None:
                return
            await self._ws.close()
            self._ws = None

    async def call(self, module, name, args, kwargs, ref_id):
        await self.connect()

        d = dict(module=module, name=name, args=args, kwargs=kwargs, ref_id=ref_id)

        # gather时会出错，只能用lock保证一次只能一个请求和响应
        async with self._lock:
            await send_in_chunks(self._ws, pickle.dumps(d), muted_print)

            file = sys.stderr
            print(f'接收数据: [', end='', file=file)
            buffer = bytearray()
            buf = bytearray()
            i = -1
            size = 0
            async for msg in self._ws:
                if msg.type is aiohttp.WSMsgType.BINARY:
                    buf.extend(msg.data)
                elif msg.type == aiohttp.WSMsgType.TEXT:
                    if msg.data == "\r\n":
                        size += len(buf)
                        buffer.extend(zlib.decompress(buf))
                        buf.clear()
                        i += 1
                        update_progress(i, print, file=file)
                    elif msg.data == "EOF":
                        print(f'] 解压完成 ({size:,}/{len(buffer):,} bytes)', file=file)
                        rsp = pickle.loads(buffer)
                        buffer.clear()
                        return rsp
                # elif msg.type is aiohttp.WSMsgType.PING:
                #     await self._ws.pong()
                # elif msg.type is aiohttp.WSMsgType.PONG:
                #     print("Pong received")
                else:
                    if msg.type is aiohttp.WSMsgType.CLOSE:
                        print('Client WebSocket connection close')
                        await self._ws.close()
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        print('Client WebSocket connection closed with exception %s' % self._ws.exception())
                    elif msg.type is aiohttp.WSMsgType.CLOSED:
                        print('Client WebSocket connection closed')

                    break

        # TODO 服务端忽然中断，客户端会直接到这一步
        # 这会让ddump开始retry
        raise RuntimeError("End of call")



import multiprocessing
import sys
import time
from typing import Dict, Any

from IPython.display import clear_output

from ksrpc.utils.process import run_command, ProcessManager


def callback(process_name, is_stderr, line, shared_time, shared_count, clear_count):
    # 使用控制台输出清屏，但异步发送数据时，进度条直接显示完成了。所以要多留一些空闲时间
    with shared_time.get_lock():
        shared_time.value = time.perf_counter()
    with shared_count.get_lock():
        i = shared_count.value + 1
        shared_count.value = i

    # 定义颜色代码
    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    GREEN = "\033[32m"

    # 根据流类型设置颜色
    if is_stderr:
        color = RED
    else:
        color = GREEN

    # 创建带颜色的输出
    colored_line = f"{BOLD}{color}[{process_name}]{RESET}: {line}"

    # 每N条清屏一次
    if i % clear_count == 0:
        clear_output(wait=True)

    if is_stderr:
        print(line, file=sys.stderr)
    else:
        print(colored_line, file=sys.stdout)


def main(commands: Dict[str, Any], idle_timeout: int = 60 * 10, clear_count: int = 40):
    """

    Parameters
    ----------
    commands:
        命令
    idle_timeout: int
        空闲多少秒后停止服务。利用的控制台输出，发送大数据时，发送进度条已经走完，但数据还在发送中，过短可能导致服务中断
    clear_count: int
        输出多少行后清屏

    """
    print('注意：子进程模式下，通过管道获取标准输出，进度条是一次性打印', file=sys.stderr)
    # 创建共享变量 (使用'd'表示双精度浮点数)
    shared_time = multiprocessing.Value('d', time.perf_counter())
    shared_count = multiprocessing.Value('i', 0)

    processes = []
    for k, v in commands.items():
        p = multiprocessing.Process(target=run_command, name=k, args=(v, callback, shared_time, shared_count, clear_count))
        processes.append(p)

    try:
        print(f"!!!注意：空闲 {idle_timeout} 秒后，应用将退出!!!", file=sys.stderr)
        with ProcessManager(*processes):
            while True:
                current_time = time.perf_counter()
                # 读取共享变量值
                with shared_time.get_lock():
                    last_activity = shared_time.value

                print(f"最后活动时间: {last_activity:.0f}, "
                      f"当前时间: {current_time:.0f}, "
                      f"空闲时间: {current_time - last_activity:-3.0f}/{idle_timeout}")

                if current_time - last_activity > idle_timeout:
                    print("!!!空闲时间到，退出!!!", file=sys.stderr)
                    break
                time.sleep(15)
    except KeyboardInterrupt:
        print("主进程已中断")
    except Exception as e:
        print("主进程错误: ", e, file=sys.stderr)
    finally:
        print("结束服务")


if __name__ == '__main__':
    commands = {
        # "ksrpc": ["python", "-u", "-m", "ksrpc.run_app"],
        # "frpc": ["./frpc", "-c", "./frpc.toml"],
        "ksrpc": ["uv", "run", "python", "-u", "-m", "ksrpc.run_app", "--config", r"./config.py"],
    }
    main(commands, idle_timeout=300, clear_count=35)



"""
演示函数
"""
import asyncio
import random
import time

import numpy as np


def sync_say_hi(name):
    n = random.randint(1, 3)
    time.sleep(n)
    return f'Hello {name}'


async def async_say_hi(name):
    n = random.randint(1, 3)
    await asyncio.sleep(n)
    return f'Hello {name}'


def div(a, b):
    return a / b


def test():
    import pandas as pd
    df = pd.DataFrame({'A': [1, 2, 3, 4, 5]})
    return df


def create_1d_array(target_mb: int = 100):
    """
    创建指定内存大小的一维 NumPy 数组

    参数:
    target_mb -- 目标内存大小 (MB)，默认为 100MB
    dtype -- 数组数据类型，默认为 np.float64

    返回:
    ndarray -- 创建的 NumPy 一维数组
    """
    dtype = np.float64
    # 计算目标字节数
    target_bytes = target_mb * 1024 * 1024

    # 获取数据类型大小
    element_size = np.dtype(dtype).itemsize

    # 计算所需元素数量
    num_elements = target_bytes // element_size

    # 创建随机数组
    arr = np.random.rand(num_elements).astype(dtype)

    # 验证并打印结果
    actual_mb = arr.nbytes / (1024 * 1024)
    print(f"创建成功: {actual_mb:.2f} MB 一维数组")
    print(f"数据类型: {arr.dtype}")
    print(f"元素数量: {arr.size:,}")

    return arr


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __format__(self, format_spec):
        # 如果没有提供格式说明符或为空，使用默认表示
        if format_spec == "":
            return f"({self.x}, {self.y})"
        # 根据不同的格式说明符返回不同的格式
        elif format_spec == "c":
            return f"({self.x}, {self.y})"
        elif format_spec == "p":
            return f"Point({self.x}, {self.y})"
        elif format_spec == "r":  # 例如，保留两位小数
            return f"({round(self.x, 2)}, {round(self.y, 2)})"
        else:
            # 对于无法识别的格式说明符，可以返回默认表示或抛出异常
            return f"({self.x}, {self.y})"


# 使用示例
p = Point(3.14159, 2.71828)


def test_array():
    arr = create_1d_array()
    return arr[:100]


PASSWORD = 123456
LIST = [1, 2, 3, 4, 5]


class A:
    class B:
        C = [11, 22, 33, 44, 55]

    D = B


CLASS = A


def sync_counter():
    n = 1
    while n <= 3:
        yield n
        n += 1

async def async_counter():
    n = 1
    while n <= 3:
        yield n
        n += 1


import hashlib


def make_key(module, name, args, kwargs, ref_id):
    """生成缓存key"""
    key = f'{module}::{name}_{repr(args)}_{repr(kwargs)}_{ref_id}'
    return hashlib.md5(key.encode('utf-8')).hexdigest()



import asyncio
import sys
import time
import zlib

import aiohttp
import dill as pickle

from ksrpc.connections import BaseConnection
from ksrpc.utils.chunks import data_sender
from ksrpc.utils.key_ import make_key
from ksrpc.utils.tqdm import update_progress, muted_print


async def process_response(response):
    """处理HTTP响应。根据不同的响应分别处理

    Parameters
    ----------
    response: Response

    Returns
    -------
    object
    json
    csv

    """
    if response.status != 200:
        raise Exception(f'{response.status}, {await response.text()}')

    file = sys.stderr
    print(f'接收数据: [', end='', file=file)
    buffer = bytearray()
    buf = bytearray()
    i = -1
    size = 0
    async for chunk, end_of_http_chunk in response.content.iter_chunks():
        buf.extend(chunk)
        if end_of_http_chunk:
            if len(buf) == 0:
                continue
            size += len(buf)
            buffer.extend(zlib.decompress(buf))
            buf.clear()
            i += 1
            update_progress(i, print, file=file)

    print(f'] 解压完成 ({size:,}/{len(buffer):,} bytes)', file=file)
    rsp = pickle.loads(buffer)
    buffer.clear()
    return rsp


class HttpConnection(BaseConnection):
    """HTTP请求响应模式
    1. 比WebSocket协议开销较高，延迟较高
    2. 一个请求一个连接。并发时会自动建立多个连接
    """

    def __init__(self, url: str, username=None, password=None):
        super().__init__(url, username, password)
        self._client = None
        self._lock = asyncio.Lock()
        self._timeout = aiohttp.ClientTimeout(total=60)

    async def __aenter__(self):
        """异步async with"""
        return self

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """异步async with"""
        async with self._lock:
            if self._client:
                await self._client.__aexit__(exc_type, exc_val, exc_tb)

    async def connect(self):
        async with self._lock:
            if self._client is not None:
                return
            self._client = aiohttp.ClientSession(auth=self._auth, timeout=self._timeout)

    async def reset(self):
        async with self._lock:
            if self._client is None:
                return
            await self._client.close()
            self._client = None

    async def call(self, module, name, args, kwargs, ref_id):
        """调用函数

        Parameters
        ----------
        module: str
            多层模块名
        name:str
            多层模块名+最后一个函数名
        args: tuple
            函数位置参数
        kwargs: dict
            函数命名参数
        ref_id

        """
        await self.connect()

        d = dict(module=module, name=name, args=args, kwargs=kwargs, ref_id=ref_id)
        key = make_key(module, name, args, kwargs, ref_id)

        data = pickle.dumps(d)
        headers = {}

        response = await self._client.post(
            # key服务端目前没有检查，以后可能用到
            self._url.format(time=time.time(), key=key),
            data=data_sender(data, muted_print),
            headers=headers,
            # proxy="http://192.168.31.33:9000",
        )

        return await process_response(response)



import asyncio


def async_wrapper(func, *args, **kwargs):
    try:
        # 尝试使用 Python 3.7+ 的 asyncio.run()
        return asyncio.run(func(*args, **kwargs))
    except AttributeError:
        # 回退到手动事件循环管理
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(func(*args, **kwargs))
        finally:
            loop.close()



from typing import List

from loguru import logger

from ksrpc.connections import BaseConnection

__all__ = ['RpcClient', 'rpc_iterator']


class RpcClient:
    """
    RpcClient根据调用生成网络请求。
    不建议用户直接使用，请用RpcProxy

    Warnings
    --------
    先调用一到多个`__getattr__`，然后`__call__`收集`__getattr__`的结果。
    并行时`__getattr__`的调用先后不可控，所以不能在`asyncio.gather`中使用

    Notes
    -----
    __hash__ __module__ __class__ __dict__ __dir__ 与调试有关，所以请使用__getattr__('__hash__')等代替
    __str__ __repr__ 断点时RuntimeWarning，所以也使用__getattr__('__repr__')等代替

    """

    def __init__(self,
                 module: str,
                 connection: BaseConnection,
                 ref_id: id = 0,
                 reraise: bool = True,
                 names: List[str] = []
                 ):
        """初始化

        Parameters
        ----------
        module: str
            顶层模块名
        connection: Connection
            连接对象
        ref_id: int
            服务端的对象id。用于迭代器，生成器等
        reraise: bool
            重新抛出服务端异常，而不是返回字典
        names: List[str]
            前几步的方法列表
        """
        self._module = module
        self._connection = connection
        self._ref_id = ref_id
        self._reraise = reraise
        self._names = names

    def __del__(self):
        self._connection = None
        self._names = None

    def __getattr__(self, name):
        # 注意：每次都生成新对象
        return RpcClient(self._module, self._connection, self._ref_id, self._reraise, self._names + [name])

    async def __call__(self, *args, **kwargs):
        name = '.'.join(self._names)
        # print("__call__", name)
        # 排序，参数顺序统一后，排序生成key便不会浪费了
        kwargs = dict(sorted(kwargs.items()))
        is_server_raise = False
        try:
            rsp = await self._connection.call(self._module, name, args, kwargs, self._ref_id)
            ref_id = rsp['ref_id']
            if rsp['status'] != 200:
                is_server_raise = True
                if self._reraise:
                    # 报异常，可用于迭代器等。这是服务端异常，不用走连接重置
                    raise rsp['data']
                else:
                    # 直接反回错误字典，不会报异常
                    return rsp
            if ref_id != 0:
                return RpcClient(self._module, self._connection, ref_id, self._reraise, [rsp['name']])
            return rsp['data']
        except Exception as e:
            # 服务端异常，直接抛出
            if is_server_raise:
                logger.warning(f'{self._module}::{name}, server error, {repr(e)}')
                raise
            # 本地异常，需重置
            await self._connection.reset()  # 重置
            logger.warning(f'{self._module}::{name}, local error, {repr(e)}')
            raise

    def __getitem__(self, item):
        # 语法糖，让RpcClient支持[]
        return self.__getattr__("__getitem__")(item)

    # @property
    # def __doc__(self):
    #     return self.__getattr__('__doc__')

    @property
    def __format__(self, format_spe: str = ""):
        # 这里""不能省
        return self.__getattr__('__format__')

    def __next__(self):
        return self.__getattr__("__next__")

    async def __anext__(self):
        return self.__getattr__("__anext__")

    def __iter__(self):
        # 本项目全是异步，理论上本句不会被调用
        # return self.__getattr__("__iter__")
        return self

    def __aiter__(self):
        # ksrpc.server.demo::async_counter.__aiter__.__anext__
        # return self.__getattr__("__aiter__")

        # TODO 取巧办法，等更好方案
        # ksrpc.server.demo::async_counter.__anext__
        return self


async def rpc_iterator(generator):
    """封装远程的迭代器。远程的异步和同步迭代器，都统一成异步

    Parameters
    ----------
    generator

    Examples
    --------
    ```python
    # 一定要抛出异常，否则
    demo = RpcProxy('ksrpc.server.demo', conn, reraise=True)

    async for it in rpc_iterator(demo.async_counter()):
        print(it)

    async for it in rpc_iterator(demo.sync_counter()):
        print(it)
    ```

    """
    async for it in await generator:
        try:
            yield await it()
        except (StopAsyncIteration, StopIteration):
            break



import sys


def muted_print(*args, **kwargs):
    pass


def update_progress(i, print, file=sys.stderr):
    """简化版进度条"""
    j = i % 10
    if j == 0:
        print('0', end='', file=file)
    elif j == 9:
        print('\b=', end='', file=file)
    else:
        print(f'\b{j}', end='', file=file)



import multiprocessing
import subprocess
import sys
import threading
import time

import psutil


def callback(process_name, is_stderr, line, *args, **kwargs):
    # 根据流类型添加不同前缀
    if is_stderr:
        print(f"[ERROR] {process_name}: {line}", file=sys.stderr)
    else:
        print(f"[INFO] {process_name}: {line}", file=sys.stdout)


def run_command(cmds, callback, *args, **kwargs):
    """运行子程序

    TODO 目前我还没解决 print('\b')，只能整行确定后再打印
    """
    process_name = multiprocessing.current_process().name
    is_win = (sys.platform == 'win32')

    while True:
        # 子进程崩溃后可重启
        start_time = time.perf_counter()
        try:
            # 使用 subprocess 运行外部命令
            process = subprocess.Popen(
                cmds,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                # universal_newlines=True,
            )
            print(cmds, f"启动子进程(PID:{process.pid})")

            # 定义从流中读取内容并调用回调函数的辅助函数
            def read_stream(stream, is_stderr):
                while True:
                    line = stream.readline()
                    if not line:  # 遇到EOF（通常是进程结束）
                        break
                    if is_win:
                        line = line.decode('gbk', errors='replace').strip()
                    else:
                        line = line.decode('utf-8', errors='replace').strip()

                    # 调用回调函数处理内容
                    callback(process_name, is_stderr, line.strip(), *args, **kwargs)

            # 为 stdout 和 stderr 创建单独的线程进行读取
            stdout_thread = threading.Thread(target=read_stream, args=(process.stdout, False))
            stderr_thread = threading.Thread(target=read_stream, args=(process.stderr, True))

            # 启动线程
            stdout_thread.start()
            stderr_thread.start()

            # 等待进程结束
            exit_code = process.wait()

            # 等待读取线程结束（确保所有输出都被处理）
            stdout_thread.join()
            stderr_thread.join()

            # 等待进程结束
            exit_code = process.wait()
            # 如果正常退出（退出码为0），则跳出循环
            if exit_code == 0:
                print(cmds, "子进程正常退出", exit_code)
                break
            else:
                print(cmds, "子进程异常退出", exit_code, file=sys.stderr)
        except KeyboardInterrupt:
            print(cmds, "已中断")
            break
        except Exception as e:
            print(cmds, "运行失败", e, file=sys.stderr)
        finally:
            print(cmds, "已结束")
            t = time.perf_counter() - start_time
            if t < 10:
                print(cmds, f"子进程启动后 {t:.3f} 秒退出。小于10秒，结束循环，等待人工处理", file=sys.stderr)
                break
            time.sleep(2)


def kill_subprocess(pid):
    # 清除子程序的功能
    parent = psutil.Process(pid)
    for p in parent.children():
        if not p.is_running():
            continue
        print(f"Terminating subprocess: {p.name()} (PID: {p.pid})")
        p.terminate()
        p.wait(timeout=5)

        if not p.is_running():
            continue
        print(f"Force killing subprocess: {p.name()} (PID: {p.pid})", file=sys.stderr)
        p.kill()
        p.wait()


def kill_process(p):
    # 清除程序的功能
    if not p.is_alive():
        return
    print(f"Terminating process: {p.name} (PID: {p.pid})")
    p.terminate()
    p.join(timeout=5)

    if not p.is_alive():
        return
    print(f"Force killing process: {p.name} (PID: {p.pid})", file=sys.stderr)
    p.kill()
    p.join()


class ProcessManager:
    def __init__(self, *processes):
        self.processes = processes

    def __enter__(self):
        for p in self.processes:
            p.start()
            print(f"Started process: {p.name} (PID: {p.pid})")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("Terminating processes...")
        for p in self.processes:
            kill_subprocess(p.pid)
            kill_process(p)
        print("All processes terminated")
        return False  # 不抑制异常



import aiohttp


class BaseConnection:
    def __init__(self, url: str, username=None, password=None):
        self._url = url
        if username and password:
            self._auth = aiohttp.BasicAuth(login=username, password=password, encoding="utf-8")
        else:
            self._auth = None

    async def reset(self):
        ...

    async def call(self, module, name, args, kwargs, ref_id):
        ...



import importlib
import sys


def import_module_from_path(module_name, file_path):
    """
    从指定文件路径导入模块

    Args:
        module_name (str): 要导入的模块名称（自定义）
        file_path (str): Python 文件的绝对路径

    Returns:
        module: 导入的模块对象
    """
    # 根据文件路径创建模块规格
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None:
        raise ImportError(f"无法从路径 '{file_path}' 创建模块规格")
    # 根据规格创建新的模块对象
    module = importlib.util.module_from_spec(spec)
    # 将模块添加到 sys.modules 全局字典中
    sys.modules[module_name] = module
    # 执行模块代码（加载模块）
    spec.loader.exec_module(module)
    return module


def main(config=None):
    if config:
        import_module_from_path("ksrpc.config", config)

    from ksrpc.app import start_server
    from ksrpc.utils.async_ import async_wrapper

    try:
        async_wrapper(start_server)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description="ksrpc aiohttp server",
    )
    parser.add_argument("--config", type=str, help="配置文件路径", nargs="?")
    args = parser.parse_args()

    main(args.config)



import inspect
import sys
import types
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from importlib import import_module

from loguru import logger

from ksrpc.config import CALL_IN_NEW_PROCESS
from ksrpc.utils.async_ import async_wrapper

logger.remove()
logger.add(sys.stderr,
           format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level:<8}</level> | PID:{process.id} | TID:{thread.id} | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
           level="INFO", colorize=True)


def get_func(module, name):
    # ksrpc.server.demo.CLASS.B.__getitem__(3)，这种格式只能确信第一个是模块，最后一个是方法，其他都不确定
    m = import_module(module)
    for n in name.split('.'):
        if len(n) == 0:
            continue
        if n == "__call__":
            # 默认 __call__ 是传不到服务端的，除非用户主动发起 __getattr__('__call__')
            continue
        elif hasattr(m, n):
            m = getattr(m, n)
        else:
            if n in ("__iter__", "__aiter__", '__next__', '__anext__'):
                # 需要特别处理
                continue
            else:
                m = import_module(f"{m.__name__}.{n}")
    return m


async def async_call(module, name, args, kwargs, ref_id):
    """简版异步API调用。没有各种额外功能"""
    # 返回的数据包
    d = dict(status=200,  # status.HTTP_200_OK,
             datetime=datetime.now().isoformat(),  # 加查询时间，缓存中也许可以判断是否过期
             module=module,
             name=name,
             args=args,
             kwargs=kwargs,
             ref_id=0)
    try:
        # 转成字符串，后面可能于做cache的key
        logger.info(f'{module}::{name}\t{args}\t{kwargs}'[:300])

        # ksrpc.server.demo::async_counter 这里产生的generator，ref_id传到了RpcClient
        # ksrpc.server.demo::async_counter.__aiter__.__anext__ 居然这里将错就错，用上了上次的对象 TODO 这里以后一定要改

        func = get_func(module, name)

        if name.endswith(("__next__", "__anext__")):
            # 不管模块了，直接引用全局变量中的对象
            try:
                ref = globals()[ref_id]
                # 兼容同步和异步
                if hasattr(ref, "__anext__"):
                    output = await ref.__anext__()
                else:
                    output = ref.__next__()
            except (StopIteration, StopAsyncIteration):
                # 迭代完成，移出
                globals().pop(ref_id, None)
                # StopIteration强制转换成异步
                raise StopAsyncIteration()
            except KeyError:
                raise StopAsyncIteration()
        else:
            if name.endswith("__func__"):  # isinstance(func, types.FunctionType) # 不行
                # print(ksrpc.server.demo.p.__format__.__func__)
                output = func
            elif isinstance(func, type):
                # print(ksrpc.server.demo.p.__class__)
                output = func
            # 可以调用的属性
            elif callable(func) or name.endswith("__call__"):
                # __call__不合语法，但与本地报错效果一致了
                if inspect.iscoroutinefunction(func):
                    output = await func(*args, **kwargs)
                else:
                    output = func(*args, **kwargs)
            else:
                output = func

        if isinstance(output, (types.GeneratorType, types.AsyncGeneratorType)):
            # 处理生成器。需要传引用ref_id,提供给__next__使用
            ref_id = id(output)
            globals()[ref_id] = output  # 一定要记得释放
            d['ref_id'] = ref_id
            output = None  # 生成器不可序列化，传空过去

        d['type'] = type(output).__name__
        d['data'] = output
    except Exception as e:
        d['status'] = 500  # status.HTTP_500_INTERNAL_SERVER_ERROR
        d['type'] = type(e).__name__
        d['data'] = e  # 是否有无法序列化的异常?

    return d


async def process_call(module, name, args, kwargs, ref_id):
    """进程版API调用。结束后进程退出自动释放内存

    Notes
    =====
    启动了一个新进程速度显著下降，但与网络传输大数据相比可以忽略不计。
    获得好处是释放了内存，适合内存有限的平台

    """
    with ProcessPoolExecutor(max_workers=1) as executor:
        future = executor.submit(async_wrapper, async_call, module, name, args, kwargs, ref_id)
        return future.result()


async def switch_call(module, name, args, kwargs, ref_id):
    if CALL_IN_NEW_PROCESS:
        return await process_call(module, name, args, kwargs, ref_id)
    else:
        return await async_call(module, name, args, kwargs, ref_id)



import asyncio
import base64
import time
import zlib

import dill as pickle
from aiohttp import web

from ksrpc.caller import switch_call
from ksrpc.config import USER_CREDENTIALS, HOST, PORT, PATH_HTTP, PATH_WS
from ksrpc.utils.chunks import send_in_chunks, data_sender
from ksrpc.utils.key_ import make_key


async def handle(request: web.Request) -> web.StreamResponse:
    buffer = bytearray()
    buf = bytearray()
    async for chunk, end_of_http_chunk in request.content.iter_chunks():
        buf.extend(chunk)
        if end_of_http_chunk:
            if len(buf) == 0:
                continue
            buffer.extend(zlib.decompress(buf))
            buf.clear()

    d = pickle.loads(buffer)
    key = make_key(**d)

    # TOOD 检查key的功能暂时屏蔽
    # if request.match_info.get("key", "") != key:
    #     return web.HTTPForbidden()

    data = await switch_call(**d)
    buffer.clear()

    body = pickle.dumps(data)
    headers = {'Content-Disposition': f"{key}.pkl.chunked.zip"}

    del data
    return web.Response(body=data_sender(body, print), headers=headers)


async def websocket_handler(request: web.Request) -> web.StreamResponse:
    # print(request.url)

    ws = web.WebSocketResponse()
    await ws.prepare(request)

    buffer = bytearray()
    async for msg in ws:
        if msg.type is web.WSMsgType.BINARY:
            buffer.extend(zlib.decompress(msg.data))
        elif msg.type == web.WSMsgType.TEXT:
            if msg.data == "EOF":
                data = await switch_call(**pickle.loads(buffer))
                buffer.clear()
                await send_in_chunks(ws, pickle.dumps(data), print)
                del data
        elif msg.type == web.WSMsgType.ERROR:
            print('Server WebSocket connection closed with exception %s' % ws.exception())
        elif msg.type is web.WSMsgType.CLOSE:
            print('Server WebSocket connection closed')
            break
    print("End of websocket_handler")
    return ws


@web.middleware
async def basic_auth_middleware(request, handler):
    # 检查是否需要跳过认证的路由
    if request.path == '/public':
        return await handler(request)

    # 获取 Authorization 头
    auth_header = request.headers.get('Authorization')

    if auth_header is None or not auth_header.startswith('Basic '):
        return unauthorized_response()

    # 解码凭证
    try:
        auth_bytes = base64.b64decode(auth_header[6:])
        auth_str = auth_bytes.decode('utf-8')
        username, password = auth_str.split(':', 1)
    except (ValueError, UnicodeDecodeError):
        return unauthorized_response()

    # 验证凭证
    if USER_CREDENTIALS.get(username) != password:
        return unauthorized_response()

    # 认证通过，添加用户名到请求对象
    request['user'] = username
    return await handler(request)


def unauthorized_response():
    return web.Response(
        status=401,
        headers={'WWW-Authenticate': 'Basic realm="Restricted Area"'},
        text="Unauthorized"
    )


@web.middleware
async def url_check_middleware(request, handler):
    # URL动态变化，防止重放攻击
    t1 = time.time()
    t2 = float(request.match_info.get("time", "0"))
    if abs(t1 - t2) > 15:
        print("HTTPForbidden:", t1, t2, t1 - t2)
        return web.HTTPForbidden()

    return await handler(request)


def create_app(argv):
    # uv run python -m aiohttp.web -H 0.0.0.0 -P 8080 ksrpc.run_app:sync_app
    app = web.Application(middlewares=[
        url_check_middleware,  # 时间检验
        basic_auth_middleware,  # 注释此行屏蔽Baisc认证
    ])
    app.add_routes([
        # TODO 路径按需修改，更安全
        web.post(PATH_HTTP, handle),
        web.get(PATH_WS, websocket_handler),
    ])
    return app


async def start_server():
    app = create_app([])
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, HOST, PORT)  # 可选：指定端口
    await site.start()
    print(f"Server started at http://{HOST}:{PORT}")
    # 保持服务器运行，直到被中断
    try:
        await asyncio.Future()  # 永久等待
    finally:
        print("Cleanup server")
        await runner.cleanup()



import os

"""
Web服务配置。不建议使用默认值
"""
HOST = "0.0.0.0"
PORT = 8080
PATH_HTTP = "/api/v1/{time}"  # HTTP服务向外提供路径。{time}表示时间动态URL，与服务器误差15秒内才能访问
PATH_WS = "/ws/v1/{time}"  # WebSocket服务向外提供路径。{time}表示时间动态URL，与服务器误差15秒内才能访问

"""
TODO Basic认证。一定要修改账号，不要使用默认值
"""
USER_CREDENTIALS = {
    "用户名": "密码",
    "admin": "password123",
    "user": "secret",
}

"""
TODO 在新进程中调用开关，可以手工设置为True/False

启动新进程会消耗一点时间，但相对网络传输大数据可以忽略不计
获得的好处是新进程会自动退出释放内存减少崩溃，适合在云服务器中运行
"""
CALL_IN_NEW_PROCESS = hasattr(os, 'fork')
print(f"CALL_IN_NEW_PROCESS = {CALL_IN_NEW_PROCESS}")
